import os
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"

from . import utils
from .spec_generator_pix import SpecGeneratorPix
import numpy as np
import collections
from multiprocessing import Pool
from matplotlib import pyplot as plt



SpecGeneratorPixInit = collections.namedtuple('SpecGeneratorPixInit', ['d_t', 'xb', 'yb', 'zb', 'logage_grid',
                                                                       'metal_grid', 'alpha_grid', 'x_edges', 'y_edges',
                                                                       'use_losvd', 'use_extinc', 'extinc_factor',
                                                                       'single_alpha', 'templatesOversampled_shape',
                                                                       'new_wave', 'wave_oversampled', 'sig_oversampled', 'factor', 'logger',
                                                                       'templatesTransposed', 'interpolator_method'])
SpecGeneratorPixInit.__new__.__defaults__ = (None,) * len(SpecGeneratorPixInit._fields)
SpecGeneratorPixCall = collections.namedtuple('SpecGeneratorPixCall', ['i', 'j'])

init_arguments = SpecGeneratorPixInit()




def init_worker(parameters, *args):

    global init_arguments

    assert not args
    assert isinstance(parameters, SpecGeneratorPixInit)

    init_arguments = parameters



def parallel_specgeneratorpix(pix_parameters):

    global init_arguments

    assert isinstance(pix_parameters, SpecGeneratorPixCall)

    specgeneratorpix = SpecGeneratorPix(d_t=init_arguments.d_t,
                                        xb=init_arguments.xb,
                                        yb=init_arguments.yb,
                                        zb=init_arguments.zb,
                                        logage_grid=init_arguments.logage_grid,
                                        metal_grid=init_arguments.metal_grid,
                                        alpha_grid=init_arguments.alpha_grid,
                                        x_edges=init_arguments.x_edges,
                                        y_edges=init_arguments.y_edges,
                                        use_losvd=init_arguments.use_losvd,
                                        use_extinc=init_arguments.use_extinc,
                                        extinc_factor=init_arguments.extinc_factor,
                                        single_alpha=init_arguments.single_alpha,
                                        templatesOversampled_shape=init_arguments.templatesOversampled_shape,
                                        new_wave=init_arguments.new_wave,
                                        wave_oversampled=init_arguments.wave_oversampled,
                                        sig_oversampled=init_arguments.sig_oversampled,
                                        factor=init_arguments.factor,
                                        logger=init_arguments.logger,
                                        templatesTransposed=init_arguments.templatesTransposed,
                                        interpolator_method=init_arguments.interpolator_method,
                                        )

    result = specgeneratorpix(i=pix_parameters.i,
                              j=pix_parameters.j)

    del specgeneratorpix
    return result





class Spec_Generator():

    def __init__(self, d_t, x_edges, y_edges, statistic_count, ssp_model, cube_params, ssp_params, other_params,
                 filepath, logger, cube_idx):

        self.d_t = d_t
        self.x_edges = x_edges
        self.y_edges = y_edges
        self.statistic_count = statistic_count
        self.ssp_model = ssp_model
        self.cube_params = cube_params
        self.ssp_params = ssp_params
        self.other_params = other_params
        self.filepath = filepath
        self.logger = logger
        self.cube_idx = cube_idx


    def __call__(self):

        self.logger.info('Start the processes of generating spectra using {0} CPU(s).'.format(self.other_params['ncpu']))
        self.logger.info('Initializing the MDF bin and data cube array...')

        self.mass_fraction_pixel_bin = np.zeros(list(self.ssp_model.reg_dim) + [self.x_edges.shape[0] - 1, self.y_edges.shape[0] - 1])
        self.data_cube = np.zeros([self.x_edges.shape[0] - 1, self.y_edges.shape[0] - 1, self.ssp_model.new_wave.shape[0]])

        self.logger.info('Initializing the common variables to be used in each process...')
        initial_arguments = SpecGeneratorPixInit(d_t=self.d_t,
                                                 xb=self.ssp_model.xb,
                                                 yb=self.ssp_model.yb,
                                                 zb=self.ssp_model.zb,
                                                 logage_grid=self.ssp_model.logage_grid,
                                                 metal_grid=self.ssp_model.metal_grid,
                                                 alpha_grid=self.ssp_model.alpha_grid,
                                                 x_edges=self.x_edges,
                                                 y_edges=self.y_edges,
                                                 use_losvd=self.cube_params['use_losvd'],
                                                 use_extinc=self.cube_params['use_extinc'],
                                                 extinc_factor=self.cube_params['extinc_factor'],
                                                 single_alpha=self.ssp_params['single_alpha'],
                                                 templatesOversampled_shape=self.ssp_model.templatesOversampled.shape,
                                                 new_wave=self.ssp_model.new_wave,
                                                 wave_oversampled=self.ssp_model.wave_oversampled,
                                                 sig_oversampled=self.ssp_model.sig_oversampled,
                                                 factor=self.ssp_model.factor,
                                                 logger=self.logger,
                                                 templatesTransposed=self.ssp_model.templatesTransposed,
                                                 interpolator_method=self.ssp_params['spec_interpolator'])


        self.logger.info('Sort the spatial pixel in a descending order in terms of number of particles (to speed up the processing).')
        index_array = np.dstack(np.unravel_index(np.argsort(self.statistic_count.ravel())[::-1], self.statistic_count.shape))

        self.logger.info('Prepare the spatial pixel indexes for the loop.')
        index_i = self.x_edges.shape[0] - 2 - index_array[0, :, 0]
        index_j = index_array[0, :, 1]

        pool = Pool(processes=self.other_params['ncpu'], initializer=init_worker, initargs=[initial_arguments])

        self.logger.info('Start looping each spatial pixel to generate the spectra...')

        generator_input = []

        for i, j in zip(index_i, index_j):

                if self.statistic_count[self.x_edges.shape[0] - 2 - i, j]!=0:

                    generator_input.append(SpecGeneratorPixCall(i=i, j=j))

                if (np.mod(len(generator_input), self.other_params['nprocess_ploop'])==0 and len(generator_input)!=0) or (i==self.x_edges.shape[0]-2 and j==self.y_edges.shape[0]-2):

                    self.logger.info('Allocate processes to generate the spectra for {0} spatial pixels.'.format(len(generator_input)))

                    _results = pool.map_async(parallel_specgeneratorpix, generator_input)
                    results = _results.get()
                    for result in results:
                        self.mass_fraction_pixel_bin[:, :, :, result.i, result.j] = result.mass_fraction_pixel
                        self.data_cube[result.i, result.j, :] = result.galaxy_rebin

                    self.logger.info('Finish one CPU loop, initialize the input list for another loop.')
                    generator_input = []

        if len(generator_input) > 0:
            _results = pool.map_async(parallel_specgeneratorpix, generator_input)
            results = _results.get()
            for result in results:
                self.mass_fraction_pixel_bin[:, :, :, result.i, result.j] = result.mass_fraction_pixel
                self.data_cube[result.i, result.j, :] = result.galaxy_rebin
        self.logger.info('Finish the last loop, initialize the input list for another loop.')


        pool.close()
        pool.join()

        self.logger.info('All the spectra generating process has been finished.')

        self.mass_fraction_pixel_bin = np.transpose(self.mass_fraction_pixel_bin, (0, 1, 2, 4, 3))
        np.save(self.filepath + 'mass_fraction_array_' + str(self.cube_idx) + '.npy', self.mass_fraction_pixel_bin)
        self.logger.info('Save the mass fraction distribution (MFD) of each spatial bin in the folder.')


        self.logger.info('Plot the statictial distribution using MFD grids and save it in the folder.')
        utils.plot_parameter_maps(self.mass_fraction_pixel_bin, self.ssp_model.age_grid_2d, self.ssp_model.metal_grid_2d,
                                  self.ssp_model.alpha_grid, self.ssp_model.reg_dim, self.x_edges, self.y_edges,
                                  xlabel=self.cube_params['x_coord'], ylabel=self.cube_params['y_coord'], cmap=plt.cm.Spectral_r)
        plt.savefig(self.filepath + 'grids_distrib_' + str(self.cube_idx) + '.png', dpi=150)





from . import utils
import numpy as np
from astropy.table import Table
from scipy.stats import binned_statistic_2d
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle



def spatial_binner(d_t, cube_params, other_params, age_grid, metal_grid, alpha_grid,
                   filepath, logger, input_arg, galaxy_dist, nparticles):
    '''
    Preparing the stellar particle catalog, N_particle spatial distribution,
    x and y axis edges from the loaded particle catalog and data cube configurations.
    :param d_t: stellar particle catalog
    :param cube_params: cube configurations
    :param other_params: other configurations
    :param age_grid: age grid of SSP templates
    :param metal_grid: metallicity grid of SSP templates
    :param alpha_grid: alpha grid of SSP templates
    :param filepath: outputs file path of this run
    :param logger: log file
    :param input_arg: setup input cube name path
    :param galaxy_dist: galaxy distance to the observer
    :param nparticles: number of particles in the catalog
    :return: list of stellar particle catalog, N_particle spatial distribution,
    x and y axis edges to be used for each data cube run.
    '''

    spatial_resolution = cube_params['spatial_resolution']
    spatial_resolution_deg = np.array(spatial_resolution) / 3600.
    spatial_resolution_x = spatial_resolution_deg[0]
    spatial_resolution_y = spatial_resolution_deg[1]

    x_coord = cube_params['x_coord']
    y_coord = cube_params['y_coord']


    if cube_params['instrument'].upper() != "DEFAULT":
        logger.info('Load the cube list from %s' % (input_arg + '_list'))
        cube_centers = np.genfromtxt(input_arg + '_list', dtype=float, delimiter=',')
        if len(cube_centers.shape) == 1:
            cube_centers = np.array([cube_centers])
        logger.info('Load the list of center positions of data cubes.')
        nbin_x = cube_params['spatial_nbin'][0]
        nbin_y = cube_params['spatial_nbin'][1]
    else:
        logger.info(
            'The instrument is in %s, will generate a huge cube using all the particles.' % cube_params['instrument'].upper())
        if cube_params['spatial_percentile'] == True:
            x_edges = np.arange(np.percentile(d_t[x_coord], 0.05),
                                np.percentile(d_t[x_coord], 99.95) + spatial_resolution_x, spatial_resolution_x)
            y_edges = np.arange(np.percentile(d_t[y_coord], 0.05),
                                np.percentile(d_t[y_coord], 99.95) + spatial_resolution_y, spatial_resolution_y)
        else:
            x_edges = np.arange(np.nanmin(d_t[x_coord]), np.nanmax(d_t[x_coord]) + spatial_resolution_x,
                                spatial_resolution_x)
            y_edges = np.arange(np.nanmin(d_t[y_coord]), np.nanmax(d_t[y_coord]) + spatial_resolution_y,
                                spatial_resolution_y)
        nbin_x = x_edges.shape[0] - 1
        nbin_y = y_edges.shape[0] - 1
        cube_centers = np.array([[(x_edges[-1] + x_edges[0]) / 2, (y_edges[-1] + y_edges[0]) / 2]])



    logger.info('================Cube info=================')
    logger.info('x_coordinate:          %s' % x_coord)
    logger.info('y_coordinate:          %s' % y_coord)
    logger.info('instrument:            %s' % cube_params['instrument'])
    logger.info('Spatial resolution:    %s' % spatial_resolution)
    logger.info('n_bins in x and y:     %s' % [nbin_x, nbin_y])
    logger.info('Number of cubes:       %s' % len(cube_centers))
    logger.info('         %s          %s' % (x_coord, y_coord))
    for i in range(len(cube_centers)):
        logger.info('      %.3f      %.3f' % (cube_centers[i][0], cube_centers[i][1]))
    logger.info('use losvd?:            %s' % cube_params['use_losvd'])
    logger.info('use extinc?:           %s' % cube_params['use_extinc'])
    logger.info('extinc factor:         %s' % cube_params['extinc_factor'])
    logger.info('use dist? :            %s' % cube_params['use_dist'])
    logger.info('add noise?:            %s' % cube_params['add_noise'])
    logger.info('bootstrapping?:        %s' % cube_params['bootstrap_table'])
    logger.info('==========================================')


    if 'cube_m_h' in d_t.keys() and 'cube_alpha_fe' in d_t.keys() and 'cube_age' in d_t.keys():
        logger.info("Not create the 'cube_x', 'mass', 'dist' columns, already exist.")
        d_t['cube_m_h'] = np.clip(d_t['cube_m_h'], metal_grid[0], metal_grid[-1])
        d_t['cube_alpha_fe'] = np.clip(d_t['cube_alpha_fe'], alpha_grid[0], alpha_grid[-1])
        d_t['cube_age'] = np.clip(d_t['cube_age'], age_grid[0], age_grid[-1])
        d_t['cube_logage'] = np.log10(d_t['cube_age']) + 9
    else:
        logger.info('Make the initial mass to be 1 M_sun.')
        d_t['mass'] = np.ones(d_t['m_ini'].shape)  # Make the mass equal to 1
        # np.clip the values in ([M/H], [alpha/Fe], age)
        logger.info('Transfer from [Fe/H] to [M/H], clipping [M/H], age and [alpha/Fe].')
        d_t['m_h'] = d_t['fe_h'] + np.log10(0.684 * (10 ** d_t['alpha_fe']) + 0.306)
        d_t['cube_m_h'] = np.clip(d_t['m_h'], metal_grid[0], metal_grid[-1])
        d_t['cube_alpha_fe'] = np.clip(d_t['alpha_fe'], alpha_grid[0], alpha_grid[-1])
        d_t['cube_age'] = np.clip(d_t['age'], age_grid[0], age_grid[-1])
        d_t['cube_logage'] = np.log10(d_t['cube_age']) + 9
        
    # Subdivide mass of the particle based on its alpha value
    # This fraction is not needed, but will keep in case one day will add it back
    d_t['fraction'] = 2.5 * d_t['cube_alpha_fe']


    # Add dist to the table
    if cube_params['use_dist']==True:
        d_t['dist'] = d_t['r']
    else:
        d_t['dist'] = np.ones(d_t['r'].shape) * galaxy_dist

    # Randomly add some numbers to bin_x and bin_y
    d_t['bin_x'] = np.zeros(len(d_t['mass']))
    d_t['bin_y'] = np.zeros(len(d_t['mass']))

    x_edges_plt = np.arange(np.percentile(d_t[x_coord], 0.05),
                            np.percentile(d_t[x_coord], 99.95) + spatial_resolution_deg[0], spatial_resolution_deg[0])
    y_edges_plt = np.arange(np.percentile(d_t[y_coord], 0.05),
                            np.percentile(d_t[y_coord], 99.95) + spatial_resolution_deg[1], spatial_resolution_deg[1])

    if other_params['plot_statistics'] == True:
        logger.info('Plotting the statistical distribution using several parameters using the original particles...')
        plt.figure(figsize=[12, 10])
        plt.subplot(321)
        im, ax, num_counts, xbins, ybins = utils.plot_binned_grids_color(x=d_t[x_coord], y=d_t[y_coord], values=d_t['vr'],
                                                                 statistic='count', x_edges=x_edges_plt, y_edges=y_edges_plt,
                                                                 xlabel=r"$" + x_coord + "$" + ' (degrees)',
                                                                 ylabel=r"$" + y_coord + "$" + ' (degrees)',
                                                                 cmap=plt.cm.hot, cblabel='Count', color_Lognorm=True)
        plt.subplot(322)
        im, ax, stats_age, xbins, ybins = utils.plot_binned_grids_color(x=d_t[x_coord], y=d_t[y_coord], values=d_t['cube_age'],
                                                                statistic='median', x_edges=x_edges_plt, y_edges=y_edges_plt,
                                                                xlabel=r"$" + x_coord + "$" + ' (degrees)',
                                                                ylabel=r"$" + y_coord + "$" + ' (degrees)',
                                                                cmap=plt.cm.Spectral_r, cblabel='Age (Gyr)', color_Lognorm=False)
        plt.subplot(323)
        im, ax, stats_feh, xbins, ybins = utils.plot_binned_grids_color(x=d_t[x_coord], y=d_t[y_coord], values=d_t['cube_m_h'],
                                                                statistic='median', x_edges=x_edges_plt, y_edges=y_edges_plt,
                                                                xlabel=r"$" + x_coord + "$" + ' (degrees)',
                                                                ylabel=r"$" + y_coord + "$" + ' (degrees)',
                                                                cmap=plt.cm.Spectral_r, cblabel='[M/H]', color_Lognorm=False)
        plt.subplot(324)
        im, ax, stats_alpha, xbins, ybins = utils.plot_binned_grids_color(x=d_t[x_coord], y=d_t[y_coord], values=d_t['cube_alpha_fe'],
                                                                  statistic='median', x_edges=x_edges_plt, y_edges=y_edges_plt,
                                                                  xlabel=r"$" + x_coord + "$" + ' (degrees)',
                                                                  ylabel=r"$" + y_coord + "$" + ' (degrees)',
                                                                  cmap=plt.cm.Spectral_r, cblabel='[Alpha/Fe]', color_Lognorm=False)
        plt.subplot(325)
        im, ax, stats_vlos, xbins, ybins = utils.plot_binned_grids_color(x=d_t[x_coord], y=d_t[y_coord], values=d_t['vr'],
                                                                 statistic='median', x_edges=x_edges_plt, y_edges=y_edges_plt,
                                                                 xlabel=r"$" + x_coord + "$" + ' (degrees)',
                                                                 ylabel=r"$" + y_coord + "$" + ' (degrees)',
                                                                 cmap=plt.cm.Spectral, cblabel='V_los (km/s)', color_Lognorm=False)
        plt.subplot(326)
        im, ax, stats_vsig, xbins, ybins = utils.plot_binned_grids_color(x=d_t[x_coord], y=d_t[y_coord], values=d_t['vr'],
                                                                 statistic='std', x_edges=x_edges_plt, y_edges=y_edges_plt,
                                                                 xlabel=r"$" + x_coord + "$" + ' (degrees)',
                                                                 ylabel=r"$" + y_coord + "$" + ' (degrees)',
                                                                 cmap=plt.cm.Spectral_r, cblabel='V_std (km/s)', color_Lognorm=False)
        plt.tight_layout()
        plt.savefig(filepath + 'particles_distrib.png', dpi=150)
        logger.info('The plot has been finished and saved into the outputs folder.')



    d_t = Table(d_t)
    if cube_params['invert_b'] == True:
        d_t['b'] = -d_t['b']
        logger.info('Invert b coordinates due to the wrong inclination used.')
    if cube_params['bootstrap_table'] == True:
        d_t = d_t[np.random.randint(len(d_t), size=len(d_t))]
        logger.info('Randomly Sample the table for a bootstrapping process.')
    if other_params['mode'] == 'debug':
        d_t = d_t[np.random.randint(len(d_t), size=nparticles)]
        logger.info("Running mode is 'debug', randomly selecting %s particles." % nparticles)
        if other_params['write_table'] == True:
            logger.info('Writing the model into astropy.table')
            d_t.write(filepath + 'particle_table.fits', format='fits', overwrite=True)



    d_t_l = []
    statistic_count_l = []
    x_edges_l = []
    y_edges_l = []

    for i in range(len(cube_centers)):
        cube_center_i = cube_centers[i]
        cube_center_i_x = cube_center_i[0]
        cube_center_i_y = cube_center_i[1]
        start_x = cube_center_i_x - float(nbin_x) / 2 * spatial_resolution_x
        start_y = cube_center_i_y - float(nbin_y) / 2 * spatial_resolution_y
        end_x = cube_center_i_x + float(nbin_x) / 2 * spatial_resolution_x
        end_y = cube_center_i_y + float(nbin_y) / 2 * spatial_resolution_y
        x_edges = np.arange(start_x, end_x + spatial_resolution_x, spatial_resolution_x)
        y_edges = np.arange(start_y, end_y + spatial_resolution_y, spatial_resolution_y)

        statistic_count, x_edges, y_edges, bin_index = binned_statistic_2d(x=d_t[x_coord], y=d_t[y_coord],
                                                                           values=d_t['mass'], statistic='count',
                                                                           bins=[x_edges, y_edges],
                                                                           expand_binnumbers=True)
        d_t['bin_x'] = bin_index[0, :]
        d_t['bin_y'] = bin_index[1, :]

        d_t_i = d_t[(bin_index[0, :] != 0) & (bin_index[0, :] != x_edges.shape[0]) &
                    (bin_index[1, :] != 0) & (bin_index[1, :] != y_edges.shape[0])]

        if other_params['mode'] != 'debug':
            if cube_params['use_extinc'] == False:
                d_t_i.keep_columns(['ra', 'dec', 'l', 'b', 'vr', 'cube_m_h', 'cube_alpha_fe', 'cube_age', 'cube_logage',
                                    'mass', 'bin_x', 'bin_y', 'fraction', 'r', 'dist'])
            else:
                d_t_i.keep_columns(['ra', 'dec', 'l', 'b', 'vr', 'cube_m_h', 'cube_alpha_fe', 'cube_age', 'cube_logage',
                                    'mass', 'bin_x', 'bin_y', 'fraction', 'r', 'dist', 'exbv'])


        if other_params['write_table'] == True:
            logger.info('Write the E-Galaxia model into astropy.Table of particles %s in datacube No.%s' % (len(d_t_i), i + 1))
            d_t_i.write(filepath + 'particle_table_' + str(i) + '.fits', format='fits', overwrite=True)


        # save the statistic_count into npy array
        np.save(filepath + 'statistic_count' + str(i) + '.npy', np.flip(statistic_count.T, axis=1))
        logger.info('Write the number of stars array into file using particles %s in datacube No.%s' % (len(d_t_i), i + 1))

        d_t_l.append(d_t_i)
        statistic_count_l.append(statistic_count)
        x_edges_l.append(x_edges)
        y_edges_l.append(y_edges)


    logger.info('Plotting the distribution of the datacubes on the Galaxy.')
    plt.figure(figsize=[6, 2.5])
    ax = plt.subplot(111)
    im, ax, num_counts, xbins, ybins = utils.plot_binned_grids_color(x=d_t[x_coord], y=d_t[y_coord], values=d_t['vr'],
                                                         statistic='count', x_edges=x_edges_plt, y_edges=y_edges_plt,
                                                         xlabel=r"$"+x_coord+"$" + ' (degrees)',
                                                         ylabel=r"$"+y_coord+"$" + ' (degrees)',
                                                         cblabel='Number Count', color_Lognorm=True, alpha=0.7,
                                                         plot_cb=False)
    plt.gca().set_aspect('equal')
    ax.set_xlim(x_edges_plt[-1], x_edges_plt[0])
    cbar = plt.colorbar(im, ax=ax, aspect=15, pad=0.01)
    cbar.set_label('Number Count', fontsize=8)
    for i in range(len(x_edges_l)):
        ax.add_patch(Rectangle((x_edges_l[i][0], y_edges_l[i][0]), x_edges_l[i][-1] - x_edges_l[i][0],
                               y_edges_l[i][-1] - y_edges_l[i][0],
                               alpha=1, facecolor='none', edgecolor='slateblue', linewidth=1.6))
    plt.tight_layout(pad=0.03)
    plt.savefig(filepath + 'datacube_distrib.png', dpi=150)
    logger.info('The figure was successfully saved.')

    np.savetxt(filepath + 'cube_list', cube_centers, delimiter=',')


    return [d_t_l, statistic_count_l, x_edges_l, y_edges_l]




def spatial_binner_continue(cube_params, other_params, filepath, logger, input_arg):
    '''
    Same as function spatial_binner but in "continue" mode
    '''

    spatial_resolution = cube_params['spatial_resolution']
    spatial_resolution_deg = np.array(spatial_resolution) / 3600.
    spatial_resolution_x = spatial_resolution_deg[0]
    spatial_resolution_y = spatial_resolution_deg[1]

    x_coord = cube_params['x_coord']
    y_coord = cube_params['y_coord']


    assert cube_params['instrument'].upper() != "DEFAULT", "There has to be an instrument assigned for running the 'continue' mode."

    logger.info('Load the cube list from %s' % (filepath + 'cube_list'))
    cube_centers = np.genfromtxt(input_arg + '_list', dtype=float, delimiter=',')
    if len(cube_centers.shape) == 1:
        cube_centers = np.array([cube_centers])
    logger.info('Load the list of center positions of data cubes.')
    nbin_x = cube_params['spatial_nbin'][0]
    nbin_y = cube_params['spatial_nbin'][1]



    logger.info('================Cube info=================')
    logger.info('x_coordinate:          %s' % x_coord)
    logger.info('y_coordinate:          %s' % y_coord)
    logger.info('instrument:            %s' % cube_params['instrument'])
    logger.info('Spatial resolution:    %s' % spatial_resolution)
    logger.info('n_bins in x and y:     %s' % [nbin_x, nbin_y])
    logger.info('Number of cubes:       %s' % len(cube_centers))
    logger.info('         %s          %s' % (x_coord, y_coord))
    for i in range(len(cube_centers)):
        logger.info('      %.3f      %.3f' % (cube_centers[i][0], cube_centers[i][1]))
    logger.info('use losvd?:            %s' % cube_params['use_losvd'])
    logger.info('use dist? :            %s' % cube_params['use_dist'])
    logger.info('add noise?:            %s' % cube_params['add_noise'])
    logger.info('bootstrapping?:        %s' % cube_params['bootstrap_table'])
    logger.info('==========================================')


    d_t_l = []
    statistic_count_l = []
    x_edges_l = []
    y_edges_l = []

    for i in range(len(cube_centers)):
        cube_center_i = cube_centers[i]
        cube_center_i_x = cube_center_i[0]
        cube_center_i_y = cube_center_i[1]
        start_x = cube_center_i_x - float(nbin_x) / 2 * spatial_resolution_x
        start_y = cube_center_i_y - float(nbin_y) / 2 * spatial_resolution_y
        end_x = cube_center_i_x + float(nbin_x) / 2 * spatial_resolution_x
        end_y = cube_center_i_y + float(nbin_y) / 2 * spatial_resolution_y
        x_edges = np.arange(start_x, end_x + spatial_resolution_x, spatial_resolution_x)
        y_edges = np.arange(start_y, end_y + spatial_resolution_y, spatial_resolution_y)


        d_t_i = Table.read(filepath + 'particle_table_' + str(i) + '.fits')
        logger.info(
            'Read the E-Galaxia model table of particles %s in datacube No.%s' % (len(d_t_i), i + 1))
        statistic_count = np.flip(np.load(filepath + 'statistic_count' + str(i) + '.npy'), axis=1).T
        logger.info(
            'Read the number of stars array into file using particles %s in datacube No.%s' % (len(d_t_i), i + 1))

        d_t_l.append(d_t_i)
        statistic_count_l.append(statistic_count)
        x_edges_l.append(x_edges)
        y_edges_l.append(y_edges)


    return [d_t_l, statistic_count_l, x_edges_l, y_edges_l]






import os
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"

from . import utils
import numpy as np
import collections
from spectres import spectres

from scipy.stats import binned_statistic_dd
from scipy.interpolate import RegularGridInterpolator
from time import perf_counter as clock


SpecResult = collections.namedtuple('SpecResult', ['mass_fraction_pixel', 'galaxy_rebin', 'i', 'j'])

class SpecGeneratorPix():

    def __init__(self, d_t, xb, yb, zb, logage_grid, metal_grid, alpha_grid, x_edges, y_edges, use_losvd, use_extinc,
                 extinc_factor, single_alpha, templatesOversampled_shape, new_wave, wave_oversampled,
                 sig_oversampled, factor, logger, templatesTransposed, interpolator_method):

        self.d_t = d_t
        self.xb = xb
        self.yb = yb
        self.zb = zb
        self.logage_grid = logage_grid.copy()
        self.metal_grid = metal_grid.copy()
        self.alpha_grid = alpha_grid.copy()
        self.x_edges = x_edges
        self.y_edges = y_edges
        self.use_losvd = use_losvd
        self.use_extinc = use_extinc
        self.extinc_factor = extinc_factor
        self.single_alpha = single_alpha
        self.new_wave = new_wave.copy()
        self.wave_oversampled = wave_oversampled.copy()
        self.sig_oversampled = sig_oversampled.copy()
        self.templatesOversampled_shape = templatesOversampled_shape
        self.factor = factor
        self.logger = logger
        self.templatesTransposed = templatesTransposed.copy()
        self.interpolator_method = interpolator_method

        # Set up the spectra interpolator
        # mask_logage = self.logage_grid
        # mask_metal =
        # mask_alpha =
        # self.logage_grid_pix = self.logage_grid[mask_logage]
        # self.metal_grid_pix = self.metal_grid[mask_metal]
        # self.alpha_grid_pix = self.alpha_grid[mask_alpha]
        # self.templatesTransposed = self.templatesTransposed[:, mask_logage, mask_metal, mask_alpha]
        self.interpolator = self.setup_interpolator()



    def __call__(self, i, j):

        t = clock()

        self.d_t_pixel = self.d_t[(self.d_t['bin_x'] == self.x_edges.shape[0] - 1 - i) & (self.d_t['bin_y'] == j + 1)]
        # delattr(self, 'd_t')
        vel = np.array(self.d_t_pixel['vr'])
        mass = np.array(self.d_t_pixel['mass'])

        # Set up the spectra interpolator
        # mask_logage = self.logage_grid
        # mask_metal =
        # mask_alpha =
        # self.logage_grid_pix = self.logage_grid[mask_logage]
        # self.metal_grid_pix = self.metal_grid[mask_metal]
        # self.alpha_grid_pix = self.alpha_grid[mask_alpha]
        # self.templatesTransposed = self.templatesTransposed[:, mask_logage, mask_metal, mask_alpha]
        # self.interpolator = self.setup_interpolator()


        if len(self.d_t_pixel) > 0:


            # Calculate the weights
            if self.single_alpha:

                sample = np.transpose(np.vstack([self.d_t_pixel['cube_logage'], self.d_t_pixel['cube_m_h']]), (1, 0))

                assert self.interpolator_method != "interp_on_alpha", "For single alpha SSP templates, there is no option to choose 'interp_on_alpha' as the interpolation method."

                if self.interpolator_method == "linear":
                    statistic_mass, edges, bin_indexes = binned_statistic_dd(sample, mass, statistic='count',
                                                                             bins=[self.logage_grid, self.metal_grid],
                                                                             expand_binnumbers=True)
                    # Calculate the mass_fraction distribution
                    mass_fraction_pixel = np.zeros(self.templatesOversampled_shape[1:])
                    for k in range(len(sample)):
                        particle = sample[k]
                        weight_logage_left = 1 - (particle[0] - self.logage_grid[bin_indexes[0, k] - 1]) / (self.logage_grid[bin_indexes[0, k]] - self.logage_grid[bin_indexes[0, k] - 1])
                        weight_logage_right = 1 - weight_logage_left
                        weight_metal_left = 1 - (particle[1] - self.metal_grid[bin_indexes[1, k] - 1]) / (self.metal_grid[bin_indexes[1, k]] - self.metal_grid[bin_indexes[1, k] - 1])
                        weight_metal_right = 1 - weight_metal_left
                        weights_grid = np.meshgrid([weight_logage_left, weight_logage_right],
                                                   [weight_metal_left, weight_metal_right])
                        weights = (weights_grid[0] * weights_grid[1]).reshape(-1)
                        grids_index = np.meshgrid([bin_indexes[0, k] - 1, bin_indexes[0, k]],
                                                  [bin_indexes[1, k] - 1, bin_indexes[1, k]])
                        logage_index = grids_index[0].reshape(-1)
                        metal_index = grids_index[1].reshape(-1)
                        mass_fraction_pixel[logage_index, metal_index, 0] += weights

                elif self.interpolator_method == "nearest":
                    mass_fraction_pixel = binned_statistic_dd(sample, mass, statistic='sum', bins=[self.xb, self.yb]).statistic
                    mass_fraction_pixel = np.stack([mass_fraction_pixel], axis=-1)

                else:
                    raise ValueError("Calculate weights failed, the interpolation method has to be either 'nearest', 'linear', or 'interp_on_alpha'.")


            else:


                sample = np.transpose(np.vstack([self.d_t_pixel['cube_logage'], self.d_t_pixel['cube_m_h'], self.d_t_pixel['cube_alpha_fe']]), (1, 0))

                if self.interpolator_method == "linear":
                    statistic_mass, edges, bin_indexes = binned_statistic_dd(sample, mass, statistic='count',
                                                                             bins=[self.logage_grid, self.metal_grid, self.alpha_grid],
                                                                             expand_binnumbers=True)
                    # Calculate the mass_fraction distribution
                    mass_fraction_pixel = np.zeros(self.templatesOversampled_shape[1:])

                    for k in range(len(sample)):
                        particle = sample[k]

                        weight_logage_left = 1 - (particle[0] - self.logage_grid[bin_indexes[0, k] - 1]) / (self.logage_grid[bin_indexes[0, k]] - self.logage_grid[bin_indexes[0, k] - 1])
                        weight_logage_right = 1 - weight_logage_left
                        weight_metal_left = 1 - (particle[1] - self.metal_grid[bin_indexes[1, k] - 1]) / (self.metal_grid[bin_indexes[1, k]] - self.metal_grid[bin_indexes[1, k] - 1])
                        weight_metal_right = 1 - weight_metal_left
                        weight_alpha_left = 1 - (particle[2] - self.alpha_grid[bin_indexes[2, k] - 1]) / (self.alpha_grid[bin_indexes[2, k]] - self.alpha_grid[bin_indexes[2, k] - 1])
                        weight_alpha_right = 1 - weight_alpha_left
                        weights_grid = np.meshgrid([weight_logage_left, weight_logage_right],
                                                   [weight_metal_left, weight_metal_right],
                                                   [weight_alpha_left, weight_alpha_right])
                        weights = (weights_grid[0] * weights_grid[1] * weights_grid[2]).reshape(-1)
                        grids_index = np.meshgrid([bin_indexes[0, k] - 1, bin_indexes[0, k]],
                                                  [bin_indexes[1, k] - 1, bin_indexes[1, k]],
                                                  [bin_indexes[2, k] - 1, bin_indexes[2, k]])
                        logage_index = grids_index[0].reshape(-1)
                        metal_index = grids_index[1].reshape(-1)
                        alpha_index = grids_index[2].reshape(-1)

                        mass_fraction_pixel[logage_index, metal_index, alpha_index] += weights


                elif self.interpolator_method == "nearest":
                    mass_fraction_pixel = binned_statistic_dd(sample, mass, statistic='sum', bins=[self.xb, self.yb, self.zb]).statistic

                elif self.interpolator_method == "interp_on_alpha":
                    sample = sample[:, 0:2]
                    fraction = 2.5 * self.d_t_pixel['cube_alpha_fe']
                    statistic_mass_alpha00 = binned_statistic_dd(sample, values=np.array(mass*(1-fraction)),
                                                                 statistic='sum', bins=[self.xb, self.yb]).statistic
                    statistic_mass_alpha04 = binned_statistic_dd(sample, values=np.array(mass*fraction),
                                                                 statistic='sum', bins=[self.xb, self.yb]).statistic
                    mass_fraction_pixel = np.stack([statistic_mass_alpha00, statistic_mass_alpha04], axis=-1)

                else:
                    raise ValueError(
                        "Stacking spectra failed, the interpolation method has to be either 'nearest', 'linear', or 'interp_on_alpha'.")



            # Calculate the stacked spectrum
            result_galaxy = np.zeros(self.templatesOversampled_shape[0])

            for k in range(len(sample)):
                particle = sample[k]

                if self.interpolator_method == "nearest" or self.interpolator_method == "linear":
                    result = self.interpolator(particle)[0]

                elif self.interpolator_method == "interp_on_alpha":
                    result_alpha00 = self.interpolator[0](particle)[0]
                    result_alpha04 = self.interpolator[1](particle)[0]
                    result = result_alpha00 * (1 - fraction[k]) + result_alpha04 * fraction[k]

                result = result * mass[k] / (1e-20 * 4 * np.pi * (self.d_t_pixel[k]['dist']*3.08567758e21)**2 / (3.826e33))

                if self.use_losvd == True:
                    result = utils.doppler_shift(self.wave_oversampled, result, vel[k])

                if self.use_extinc == True:
                    result = result * utils.reddening_cal00(self.wave_oversampled, self.d_t_pixel[k]['exbv'] / self.extinc_factor)

                result_galaxy += result


            # Degrade and Rebin the spectra
            galaxy_degraded = utils.degrade_spec_ppxf(result_galaxy, None, self.sig_oversampled, gau_npix=None)[0]
            galaxy_rebin = spectres(self.new_wave, self.wave_oversampled, galaxy_degraded, fill=np.nan, verbose=False)



        else:

            mass_fraction_pixel = np.zeros(self.templatesOversampled_shape[1:])
            galaxy_rebin = np.zeros(int(self.templatesOversampled_shape[0] / self.factor))


        self.logger.info('Generated a spectrum for spatial pixel %s using %s particles, time elapsed: %.2f s' % ([int(i), int(j)], len(self.d_t_pixel), clock() - t))
        return SpecResult(mass_fraction_pixel=mass_fraction_pixel, galaxy_rebin=galaxy_rebin, i=i, j=j)




    def setup_interpolator(self):
        # Set up the spectra interpolator

        if self.single_alpha:
            assert self.interpolator_method != "interp_on_alpha", "'interp_on_alpha' is not avaliable for single alpha case."
            SSP_grids = (self.logage_grid, self.metal_grid)
            interpolator = RegularGridInterpolator(SSP_grids, self.templatesTransposed[:, :, 0, :], method=self.interpolator_method)
            self.interpolator = interpolator

        else:
            if self.interpolator_method == "nearest" or self.interpolator_method == "linear":
                SSP_grids = (self.logage_grid, self.metal_grid, self.alpha_grid)
                interpolator = RegularGridInterpolator(SSP_grids, self.templatesTransposed, method=self.interpolator_method)
                self.interpolator = interpolator
            elif self.interpolator_method == "interp_on_alpha":
                SSP_grids = (self.logage_grid, self.metal_grid)
                interpolator_alpha00 = RegularGridInterpolator(SSP_grids, self.templatesTransposed[:, :, 0, :], method="nearest")
                interpolator_alpha04 = RegularGridInterpolator(SSP_grids, self.templatesTransposed[:, :, 1, :], method="nearest")
                self.interpolator = [interpolator_alpha00, interpolator_alpha04]
            else:
                raise ValueError(
                    "Initializing the interpolator ailed, the method has to be either 'nearest', 'linear', or 'interp_on_alpha'.")

        return interpolator



from setuptools import setup, find_packages
import os

def readme():
    with open('README.md') as file:
        return(file.read())

def versionNumber():
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'GalCraft/_version.py')) as versionFile:
        return(versionFile.readlines()[-1].split()[-1].strip("\"'"))

setup(name='GalCraft',
      version=versionNumber(),
      description='GalCraft: Building integral-field spectrograph data cubes of the Galaxy',
      long_description_content_type="text/markdown",
      long_description=readme(),
      classifiers=[
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering :: Astronomy',
          'Programming Language :: Python :: 3',
          'License :: OSI Approved :: MIT License',
      ],
      url='https://github.com/purmortal/galcraft',
      author='Zixian Wang (Purmortal)',
      author_email='wang.zixian.astro@gmail.com',
      packages=find_packages(),
      package_data={
          '': ['**/*']
      },
      install_requires=[
          'astropy',
          'numpy',
          'scipy',
          'matplotlib',
          'ebfpy',
          'ephem',
          'spectres',
      ],
      python_requires='>=3.6',
      entry_points={
          'console_scripts': [
              'GalCraft        = GalCraft.MainProcess:main'
          ],
      },
      include_package_data=True,
      zip_safe=False)



import logging


class Logger:
    def __init__(self, logfile, logger=None, level=logging.INFO, mode='a'):
        self.logger = logging.getLogger(logger)
        self.logger.propagate = False
        self.logger.setLevel(level)
        fh = logging.FileHandler(logfile, mode=mode, encoding='utf-8')
        fh.setLevel(level)
        sh = logging.StreamHandler()
        sh.setLevel(level)
        formatter = logging.Formatter("%(asctime)s - %(levelname)s: %(message)s")
        fh.setFormatter(formatter)
        sh.setFormatter(formatter)
        self.logger.handlers.clear()
        self.logger.addHandler(fh)
        self.logger.addHandler(sh)
        fh.close()
        sh.close()

    def get_log(self):
        return self.logger





import numpy as np
import ephem

def observe(data,oparams):
    cotm=Cot(**oparams)
    cotm.gc2helio(data)


def equ2gal(*args,epoch='2000'):
    al = args[0]
    de = args[1]
    al = np.radians(al) 
    de = np.radians(de)
    
    g = ephem.Galactic(0.0,np.pi/2.0,epoch=str(epoch))
    [al_gp,de_gp] = g.to_radec()
    g.from_radec(0.0,np.pi/2.0)
    l_cp = g.lon.znorm

    l = l_cp-np.arctan2(np.cos(de)*np.sin(al-al_gp),np.cos(de_gp)*np.sin(de)-np.sin(de_gp)*np.cos(de)*np.cos(al-al_gp))
    b = np.arcsin(np.sin(de_gp)*np.sin(de)+np.cos(de_gp)*np.cos(de)*np.cos(al - al_gp))

    if len(args) == 2:
        l = np.degrees(l)%360.0
        b = np.degrees(b)      
        return [l,b]
    else:    
        mua = args[2]
        mud = args[3]
    
        c1 = np.sin(de_gp)*np.cos(de)-np.cos(de_gp)*np.sin(de)*np.cos(al-al_gp)
        c2 = np.cos(de_gp)*np.sin(al-al_gp)
        cosb = np.sqrt(c1*c1+c2*c2)
        mul = (c1*mua+c2*mud)/cosb
        mub = (-c2*mua+c1*mud)/cosb   

        l = np.degrees(l)%360.0
        b = np.degrees(b)      
        return [l,b,mul,mub]


def gal2equ(*args, epoch='2000'):
    l = args[0]
    b = args[1]
    l = np.radians(l) 
    b = np.radians(b)
    
    g=ephem.Galactic(0.0,np.pi/2.0,epoch=str(epoch))
    [al_gp,de_gp]=g.to_radec()
    g.from_radec(0.0,np.pi/2.0)
    l_cp=g.lon.znorm   

    al = al_gp + np.arctan2(np.cos(b) * np.sin(l_cp - l), np.cos(de_gp) * np.sin(b) - np.sin(de_gp) * np.cos(b) * np.cos(l_cp - l))   
    de=np.arcsin(np.sin(de_gp)*np.sin(b)+np.cos(de_gp)*np.cos(b)*np.cos(l_cp - l))
   
    if len(args) == 2:
        al = np.degrees(al)%360.0    
        de = np.degrees(de)      
        return [al,de,mua,mud]
    else:
        mul = args[2]
        mub = args[3]
    
        c1=np.sin(de_gp)*np.cos(de)-np.cos(de_gp)*np.sin(de)*np.cos(al-al_gp)
        c2=np.cos(de_gp)*np.sin(al-al_gp)
        cosb=np.sqrt(c1*c1+c2*c2)
        mua=(c1*mul-c2*mub)/cosb
        mud=(c2*mul+c1*mub)/cosb   

        al = np.degrees(al)%360.0    
        de = np.degrees(de)
        return [al, de, mua, mud]
        


def ecl2equ(l, b,epoch='2000'):   
   g=ephem.Ecliptic(0.0,np.pi/2.0,epoch=str(epoch))
   [al_gp,de_gp]=g.to_radec()
   g.from_radec(0.0,np.pi/2.0)
   l_cp=g.lon.znorm   
   l = np.radians(l) 
   b = np.radians(b)
   
   al = al_gp + np.arctan2(np.cos(b) * np.sin(l_cp - l), np.cos(de_gp) * np.sin(b) - np.sin(de_gp) * np.cos(b) * np.cos(l_cp - l))   
   de=np.arcsin(np.sin(de_gp)*np.sin(b)+np.cos(de_gp)*np.cos(b)*np.cos(l_cp - l))
   de = np.degrees(de)      
   al = np.degrees(al)%360.0    
   return [al,de]


def equ2ecl(al, de,epoch='2000'):   
   g=ephem.Ecliptic(0.0,np.pi/2.0,epoch=str(epoch))
   [al_gp,de_gp]=g.to_radec()
   g.from_radec(0.0,np.pi/2.0)
   l_cp=g.lon.znorm
   al = np.radians(al) 
   de = np.radians(de)

   l=l_cp-np.arctan2(np.cos(de)*np.sin(al-al_gp),np.cos(de_gp)*np.sin(de)-np.sin(de_gp)*np.cos(de)*np.cos(al-al_gp))
   b=np.arcsin(np.sin(de_gp)*np.sin(de)+np.cos(de_gp)*np.cos(de)*np.cos(al - al_gp))
   b = np.degrees(b)      
   l=np.degrees(l)%360.0   
   return [l,b]
    

def lbr2xyz(*args):
    l=np.radians(args[0]);    b=np.radians(args[1]);    r=args[2]
    px=r*np.cos(b)*np.cos(l)
    py=r*np.cos(b)*np.sin(l)
    pz=r*np.sin(b)
    if len(args) == 3:
        return [px, py, pz]
    
    vl=np.radians(args[3]);    vb=np.radians(args[4]);    vr=args[5]
    tm_00=-np.sin(l) ; tm_01=-np.sin(b)*np.cos(l) ; tm_02= np.cos(b)*np.cos(l)
    tm_10= np.cos(l) ; tm_11=-np.sin(b)*np.sin(l) ; tm_12= np.cos(b)*np.sin(l)
    tm_20= 0.0       ; tm_21= np.cos(b)           ; tm_22= np.sin(b)
    vx=vl*tm_00+vb*tm_01+vr*tm_02
    vy=vl*tm_10+vb*tm_11+vr*tm_12
    vz=vl*tm_20+vb*tm_21+vr*tm_22
    return [px, py, pz, vx, vy, vz]

def xyz2lbr(*args):
    px=args[0];    py=args[1];    pz=args[2]
    rc2=px*px+py*py
    l=np.degrees(np.arctan2(py,px))
    b=np.degrees(np.arctan(pz/np.sqrt(rc2)))
    r1=np.sqrt(rc2+pz*pz)
    if len(args) == 3:
        return l, b, r

    vx=args[3];    vy=args[4];    vz=args[5]
    r=np.sqrt(rc2+pz*pz)
    ind=np.where(r == 0.0)[0]
    r[ind]=1.0
    px=px/r
    py=py/r
    pz=pz/r
    rc=np.sqrt(px*px+py*py)
    ind=np.where(rc == 0.0)[0]
    rc[ind]=1.0
    tm_00=-py/rc; tm_01=-pz*px/rc; tm_02= rc*px/rc
    tm_10= px/rc; tm_11=-pz*py/rc; tm_12= rc*py/rc
    tm_20= 0.0  ; tm_21= rc  ; tm_22= pz
    vl=(vx*tm_00+vy*tm_10+vz*tm_20)
    vb=(vx*tm_01+vy*tm_11+vz*tm_21)
    vr=(vx*tm_02+vy*tm_12+vz*tm_22)
    return [l, b, r1, vl, vb, vr]

    
def lzr2xyz(*args):
    l=args[0]; pz=args[1]; r=args[2]
    l=np.radians(l)
    px=r*np.cos(l)
    py=r*np.sin(l)
    pz=pz*1.0
    if len(args) == 3:
        return [px, py, pz]

    vl=args[3]; vz=args[4]; vr=args[5]
    tm_00=-np.sin(l) ; tm_02= np.cos(l)
    tm_10= np.cos(l) ; tm_12= np.sin(l)
    vx=vl*tm_00+vr*tm_02
    vy=vl*tm_10+vr*tm_12
    vz=vz*1.0
    return [px, py, pz, vx, vy,vz]


def xyz2lzr(*args):
    px=args[0];    py=args[1];    pz=args[2]    
    l=np.degrees(np.arctan2(py,px))
    z=np.array(pz)
    rc1=np.sqrt(px*px+py*py)
    if len(args) == 3:
        return [l,z,rc1]
    
    vx=args[3];    vy=args[4];    vz=args[5]    
    r=np.sqrt(px*px+py*py+pz*pz)
    ind=np.where(r == 0.0)[0]
    r[ind]=1.0
    px=px/r
    py=py/r
    pz=pz/r
    rc=np.sqrt(px*px+py*py)
    ind=np.where(rc == 0.0)[0]
    rc[ind]=1.0
    tm_00=-py/rc; tm_01=0.0; tm_02= px/rc
    tm_10= px/rc; tm_11=0.0; tm_12= py/rc
    tm_20= 0.0  ; tm_21=1.0; tm_22= 0.0
    vl=(vx*tm_00+vy*tm_10+vz*tm_20)
    vz=(vx*tm_01+vy*tm_11+vz*tm_21)
    vr=(vx*tm_02+vy*tm_12+vz*tm_22)
    return [l, z, rc1, vl, vz, vr]


def rotate(th,x,y):
    """
    Transformation of coordinates under rotation of coordinate sysytem 
    by an angle th.
    Alternatively, new coordinates after rotation the body defined by (x,y) 
    by an angle th about the existing coordinate system.
    Parameters
    ----------
    th- angle in degree
    (x,y)- coordinatesto be transformed 

    Returns
    ----------
    (x,y) - coordinates of (x,y) in new rotated coordinate system.

    """
    x1=x*1.0 
    y1=y*1.0
    th=np.radians(th)
    x=x1*np.cos(th)+y1*np.sin(th)
    y=-x1*np.sin(th)+y1*np.cos(th)
    return [x,y]


class Cot():
    def  __init__(self,theta_zx=0.0,theta_yz=0.0,l=0.0,b=0.0,distance=8.0):
        """
        Transform coordinate of a  galaxies from internal 
        reference frame to observational and vice versa. 
        We assume the center of the new Galaxy to be at (l, b, s). Following 
        operations are done to move and orient the Galaxy to the desired 
        location. 
        rotate_zx(-theta_zx), rotate_yz(-theta_yz) translate_x(distance), 
        rotate_zx(b), rotate_xy(-l) 

        Parameters
        ------------
        theta_zx- inclination angle [degree], rotate around y axis z to x
        theta_yz- orientation angle [degree], rotate around x axis y to z
        l- Galactic long [degree]
        b- Galactic lat [degree]
        distance-  [kpc]

        """
        self.glon=l
        self.glat=b
        self.distance=distance
        self.theta_zx=theta_zx
        self.theta_yz=theta_yz
            
    def helio2gc(self,data):
        """
        Transform from observers reference frame to a galaxies internal 
        reference.  Without any loss of genrality 
        
        Parameters
        ----------
        data- dict of np arrays containing data to be transformed    

        Returns
        -------
        The data is modified inplace. 
        (lgc,pzgc,rcgc,vlgc,vzgc,vrcgc)
        """
        d=data
       
        if 'px' not in d:
            print(d.keys())
            if ('ra' in d) and ('l' not in d):
                if 'pmra' in d:
                    d['l'],d['b'],d['vl'],d['vb']=equ2gal(d['ra'],d['dec'],d['pmra'],d['pmdec'])
                    d['vl']=d['vl']*4.74e3*d['r']
                    d['vb']=d['vb']*4.74e3*d['r']
                else:
                    d['l'],d['b']=equ2gal(d['ra'],d['dec'])
           
            if 'l' in d:
                if 'vl' in d:
                    d['px'],d['py'],d['pz'],d['vx'],d['vy'],d['vz']=lbr2xyz(d['l'],d['b'],d['r'],d['vl'],d['vb'],d['vr'])
                else:
                    d['px'],d['py'],d['pz']=lbr2xyz(d['l'],d['b'],d['r'])
            else:
                raise RuntimeError('should have glon glat')
           
        if self.glon != 0.0:
            d['px'],d['py']=rotate(self.glon,d['px'],d['py'])
            if 'vx' in d:
                d['vx'],d['vy']=rotate(self.glon,d['vx'],d['vy'])
        if self.glat != 0.0:
            d['pz'],d['px']=rotate(-self.glat,d['pz'],d['px'])
            if 'vx' in d:
                d['vz'],d['vx']=rotate(-self.glat,d['vz'],d['vx'])
           
           
        d['px']=d['px']-self.distance
        if self.theta_yz != 0.0: 
            d['py'],d['pz']=rotate(self.theta_yz,d['py'],d['pz'])
            if 'vz' in d:
                d['vy'],d['vz']=rotate(self.theta_yz,d['vy'],d['vz'])
                
        if self.theta_zx != 0.0: 
            d['pz'],d['px']=rotate(self.theta_zx,d['pz'],d['px'])
            if 'vz' in d:
                d['vz'],d['vx']=rotate(self.theta_zx,d['vz'],d['vx'])
           
        if 'l' in d:
            if 'vz' in d:
                d['lgc'],d['pzgc'],d['rcgc'],d['vlgc'],d['vzgc'],d['vrcgc']=Cot.xyz2lzr(d['px'],d['py'],d['pz'],d['vx'],d['vy'],d['vz'])
            else:
                d['lgc'],d['pzgc'],d['rcgc']=xyz2lzr(d['px'],d['py'],d['pz'])
           
       
    def gc2helio(self, data):
       """
       Transform from a galaxies internal reference frame to observers 
       reference frame. Without any loss of genrality we 
       assume the center of the new Galaxy to be at (s,0,0). Theta is the 
       inclination of the z axis of the galaxy with that of observers.

       Rotate around y axis by angle theta (z to x) a Galaxy, then  
       translate in x by distance s.
       
       Parameters
       ----------
       data- dict of np arrays containing data to be transformed    
       should contain ['lgc','pzgc','rcgc'] and optionally 
       ['vlgc','vzgc','vrcgc']
       Returns
       -------
       The data is modified inplace. 
       (l, b, r, vl, vb, vr)
       (ra, dec, pmra, pmdec)
       """
       d=data
       if 'lgc' in d:
           if 'vlgc' in d:
               d['px'],d['py'],d['pz'],d['vx'],d['vy'],d['vz']=lzr2xyz(d['lgc'],d['pzgc'],d['rcgc'],d['vlgc'],d['vzgc'],d['vrcgc'])
           else:
               d['px'],d['py'],d['pz']=lzr2xyz(d['lgc'],d['pzgc'],d['rcgc'])
       else:
           raise RuntimeError('should have lgc ')

       if self.theta_zx != 0.0: 
           d['pz'],d['px']=rotate(-self.theta_zx,d['pz'],d['px'])
           if 'vz' in d:
               d['vz'],d['vx']=rotate(-self.theta_zx,d['vz'],d['vx'])
       if self.theta_yz != 0.0: 
           d['py'],d['pz']=rotate(-self.theta_yz,d['py'],d['pz'])
           if 'vz' in d:
               d['vy'],d['vz']=rotate(-self.theta_yz,d['vy'],d['vz'])
       d['px']=d['px']+self.distance

       if self.glat != 0.0:
           d['pz'],d['px']=rotate(self.glat,d['pz'],d['px'])
           if 'vx' in d:
               d['vz'],d['vx']=rotate(self.glat,d['vz'],d['vx'])
           
       if self.glon != 0.0:
           d['px'],d['py']=rotate(-self.glon,d['px'],d['py'])
           if 'vx' in d:
               d['vx'],d['vy']=rotate(-self.glon,d['vx'],d['vy'])
       
       if 'vz' in d:
           d['l'],d['b'],d['r'],d['vl'],d['vb'],d['vr']=xyz2lbr(d['px'],d['py'],d['pz'],d['vx'],d['vy'],d['vz'])
           d['ra'],d['dec'],d['pmra'],d['pmdec']=gal2equ(d['l'],d['b'],d['vl']/(4.74e3*d['r']),d['vb']/(4.74e3*d['r']))
       else:
           d['l'],d['b'],d['r']=xyz2lbr(d['px'],d['py'],d['pz'])
           d['ra'],d['dec']=gal2equ(d['l'],d['b'])


           
